from .vertex_attribute_names import *
from .upload_file import *
from .texture_coordinates import *
from .save_viewable_file import *
from .polyhedron_attribute_names import *
from .polygon_attribute_names import *
from .ping import *
from .missing_files import *
from .kill import *
from .inspect_file import *
from .import_project import *
from .geographic_coordinate_systems import *
from .geode_objects_and_output_extensions import *
from .geode_object_inheritance import *
from .export_project import *
from .edge_attribute_names import *
from .cell_attribute_names import *
from .allowed_objects import *
from .allowed_files import *



# Standard library imports
import os

# Third party imports
from werkzeug.datastructures import FileStorage
from flask.testing import FlaskClient
from werkzeug.test import TestResponse
from pathlib import Path
import json
import zipfile

# Local application imports
from opengeodeweb_microservice.database.data import Data
from opengeodeweb_microservice.database.connection import get_session
from opengeodeweb_back import geode_functions, test_utils
from opengeodeweb_back.geode_objects.geode_polygonal_surface3d import (
    GeodePolygonalSurface3D,
)
from opengeodeweb_back.geode_objects.geode_polyhedral_solid3d import (
    GeodePolyhedralSolid3D,
)

from opengeodeweb_back.geode_objects.geode_regular_grid2d import (
    GeodeRegularGrid2D,
)
from opengeodeweb_back.geode_objects.geode_edged_curve3d import (
    GeodeEdgedCurve3D,
)

base_dir = os.path.abspath(os.path.dirname(__file__))
data_dir = os.path.join(base_dir, "data")


def test_allowed_files(client: FlaskClient) -> None:
    route = f"/opengeodeweb_back/allowed_files"

    def get_full_data() -> test_utils.JsonData:
        return {}

    json = get_full_data()
    response = client.post(route, json=json)
    assert response.status_code == 200
    extensions = response.get_json()["extensions"]
    assert type(extensions) is list
    for extension in extensions:
        assert type(extension) is str

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_allowed_objects(client: FlaskClient) -> None:
    route = f"/opengeodeweb_back/allowed_objects"

    def get_full_data() -> test_utils.JsonData:
        return {
            "filename": "corbi.og_brep",
        }

    # Normal test with filename 'corbi.og_brep'
    response = client.post(route, json=get_full_data())
    assert response.status_code == 200
    allowed_objects = response.get_json()["allowed_objects"]
    assert type(allowed_objects) is dict
    for allowed_object in allowed_objects:
        assert type(allowed_object) is str

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_upload_file(client: FlaskClient, filename: str = "test.og_brep") -> None:
    file = os.path.join(data_dir, filename)
    print(f"{file=}", flush=True)
    response = client.put(
        f"/opengeodeweb_back/upload_file",
        data={"file": FileStorage(open(file, "rb"))},
    )
    assert response.status_code == 201


def test_missing_files(client: FlaskClient) -> None:
    route = f"/opengeodeweb_back/missing_files"

    def get_full_data() -> test_utils.JsonData:
        return {
            "geode_object_type": "BRep",
            "filename": "test.og_brep",
        }

    json = get_full_data()
    response = client.post(route, json=json)
    assert response.status_code == 200
    has_missing_files = response.get_json()["has_missing_files"]
    mandatory_files = response.get_json()["mandatory_files"]
    additional_files = response.get_json()["additional_files"]
    assert type(has_missing_files) is bool
    assert type(mandatory_files) is list
    assert type(additional_files) is list

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_geographic_coordinate_systems(client: FlaskClient) -> None:
    route = f"/opengeodeweb_back/geographic_coordinate_systems"

    def get_full_data() -> test_utils.JsonData:
        return {
            "geode_object_type": "BRep",
        }

    response = client.post(route, json=get_full_data())
    assert response.status_code == 200
    crs_list = response.get_json()["crs_list"]
    assert type(crs_list) is list
    for crs in crs_list:
        assert type(crs) is dict

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_inspect_file(client: FlaskClient) -> None:
    route = f"/opengeodeweb_back/inspect_file"

    def get_full_data() -> test_utils.JsonData:
        return {
            "geode_object_type": "BRep",
            "filename": "corbi.og_brep",
        }

    json = get_full_data()

    # Normal test with geode_object 'BRep'
    response = client.post(route, json=json)
    assert response.status_code == 200
    inspection_result = response.get_json()["inspection_result"]
    assert type(inspection_result) is dict

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_geode_objects_and_output_extensions(client: FlaskClient) -> None:
    route = "/opengeodeweb_back/geode_objects_and_output_extensions"

    def get_full_data() -> test_utils.JsonData:
        return {
            "geode_object_type": "BRep",
            "filename": "corbi.og_brep",
        }

    response = client.post(route, json=get_full_data())

    assert response.status_code == 200
    geode_objects_and_output_extensions = response.get_json()[
        "geode_objects_and_output_extensions"
    ]
    assert type(geode_objects_and_output_extensions) is dict
    for geode_object, values in geode_objects_and_output_extensions.items():
        assert type(values) is dict
        for output_extension, value in values.items():
            assert type(value) is bool

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_save_viewable_file(
    client: FlaskClient,
    geode_object_type: str = "BRep",
    filename: str = "corbi.og_brep",
) -> TestResponse:
    test_upload_file(client, filename)
    route = f"/opengeodeweb_back/save_viewable_file"

    def get_full_data() -> test_utils.JsonData:
        return {
            "geode_object_type": geode_object_type,
            "filename": filename,
        }

    # Normal test with filename 'corbi.og_brep'
    response = client.post(route, json=get_full_data())
    assert response.status_code == 200
    native_file = response.get_json()["native_file"]
    assert type(native_file) is str
    viewable_file = response.get_json()["viewable_file"]
    assert type(viewable_file) is str
    id = response.get_json().get("id")
    assert type(id) is str
    object_type = response.get_json()["viewer_type"]
    assert type(object_type) is str
    assert object_type in ["model", "mesh"]
    binary_light_viewable = response.get_json()["binary_light_viewable"]
    assert type(binary_light_viewable) is str

    # Test all params
    test_utils.test_route_wrong_params(client, route, get_full_data)
    return response


def test_texture_coordinates(client: FlaskClient, test_id: str) -> None:
    with client.application.app_context():
        file = os.path.join(data_dir, "hat.vtp")
        data = Data.create(
            geode_object=GeodePolygonalSurface3D.geode_object_type(),
            viewer_object=GeodePolygonalSurface3D.viewer_type(),
            viewer_elements_type=GeodePolygonalSurface3D.viewer_elements_type(),
        )
        data.native_file = file
        session = get_session()
        if session:
            session.commit()

        data_path = geode_functions.data_file_path(data.id, data.native_file)
        os.makedirs(os.path.dirname(data_path), exist_ok=True)
        assert os.path.exists(data_path), f"File not found at {data_path}"
    response = client.post(
        "/opengeodeweb_back/texture_coordinates", json={"id": data.id}
    )
    assert response.status_code == 200
    texture_coordinates = response.get_json()["texture_coordinates"]
    assert type(texture_coordinates) is list
    for texture_coordinate in texture_coordinates:
        assert type(texture_coordinate) is str


def test_vertex_attribute_names(client: FlaskClient, test_id: str) -> None:
    route = f"/opengeodeweb_back/vertex_attribute_names"

    with client.application.app_context():
        file = os.path.join(data_dir, "test.vtp")
        data = Data.create(
            geode_object=GeodePolygonalSurface3D.geode_object_type(),
            viewer_object=GeodePolygonalSurface3D.viewer_type(),
            viewer_elements_type=GeodePolygonalSurface3D.viewer_elements_type(),
        )
        data.native_file = file
        session = get_session()
        if session:
            session.commit()

        data_path = geode_functions.data_file_path(data.id, data.native_file)
        os.makedirs(os.path.dirname(data_path), exist_ok=True)
        assert os.path.exists(data_path), f"File not found at {data_path}"
    response = client.post(route, json={"id": data.id})
    assert response.status_code == 200
    attributes = response.get_json()["attributes"]
    assert type(attributes) is list
    for attribute in attributes:
        assert "attribute_name" in attribute
        assert "min_value" in attribute
        assert "max_value" in attribute


def test_cell_attribute_names(client: FlaskClient, test_id: str) -> None:
    route = f"/opengeodeweb_back/cell_attribute_names"

    with client.application.app_context():
        file = os.path.join(data_dir, "test.og_rgd2d")
        data = Data.create(
            geode_object=GeodeRegularGrid2D.geode_object_type(),
            viewer_object=GeodeRegularGrid2D.viewer_type(),
            viewer_elements_type=GeodeRegularGrid2D.viewer_elements_type(),
        )
        data.native_file = file
        session = get_session()
        if session:
            session.commit()

        data_path = geode_functions.data_file_path(data.id, data.native_file)
        os.makedirs(os.path.dirname(data_path), exist_ok=True)
        assert os.path.exists(data_path), f"File not found at {data_path}"
    response = client.post(route, json={"id": data.id})
    assert response.status_code == 200
    attributes = response.get_json()["attributes"]
    assert type(attributes) is list
    for attribute in attributes:
        assert "attribute_name" in attribute
        assert "min_value" in attribute
        assert "max_value" in attribute


def test_polygon_attribute_names(client: FlaskClient, test_id: str) -> None:
    route = f"/opengeodeweb_back/polygon_attribute_names"

    with client.application.app_context():
        file = os.path.join(data_dir, "test.vtp")
        data = Data.create(
            geode_object=GeodePolygonalSurface3D.geode_object_type(),
            viewer_object=GeodePolygonalSurface3D.viewer_type(),
            viewer_elements_type=GeodePolygonalSurface3D.viewer_elements_type(),
        )
        data.native_file = file
        session = get_session()
        if session:
            session.commit()

        data_path = geode_functions.data_file_path(data.id, data.native_file)
        os.makedirs(os.path.dirname(data_path), exist_ok=True)
        assert os.path.exists(data_path), f"File not found at {data_path}"
    response = client.post(route, json={"id": data.id})
    assert response.status_code == 200
    attributes = response.get_json()["attributes"]
    assert type(attributes) is list
    for attribute in attributes:
        assert "attribute_name" in attribute
        assert "min_value" in attribute
        assert "max_value" in attribute


def test_polyhedron_attribute_names(client: FlaskClient, test_id: str) -> None:
    route = f"/opengeodeweb_back/polyhedron_attribute_names"

    with client.application.app_context():
        file = os.path.join(data_dir, "test.vtu")
        data = Data.create(
            geode_object=GeodePolyhedralSolid3D.geode_object_type(),
            viewer_object=GeodePolyhedralSolid3D.viewer_type(),
            viewer_elements_type=GeodePolyhedralSolid3D.viewer_elements_type(),
        )
        data.native_file = file
        session = get_session()
        if session:
            session.commit()

        data_path = geode_functions.data_file_path(data.id, data.native_file)
        os.makedirs(os.path.dirname(data_path), exist_ok=True)
        assert os.path.exists(data_path), f"File not found at {data_path}"
    response = client.post(route, json={"id": data.id})
    print(response.get_json())
    assert response.status_code == 200
    attributes = response.get_json()["attributes"]
    assert type(attributes) is list
    for attribute in attributes:
        assert "attribute_name" in attribute
        assert "min_value" in attribute
        assert "max_value" in attribute
        if attribute["attribute_name"] == "Range":
            assert attribute["min_value"] == 0.0
            assert attribute["max_value"] == 579.0


def test_edge_attribute_names(client: FlaskClient, test_id: str) -> None:
    route = f"/opengeodeweb_back/edge_attribute_names"

    with client.application.app_context():
        file = os.path.join(data_dir, "test.og_edc3d")
        data = Data.create(
            geode_object=GeodeEdgedCurve3D.geode_object_type(),
            viewer_object=GeodeEdgedCurve3D.viewer_type(),
            viewer_elements_type=GeodeEdgedCurve3D.viewer_elements_type(),
        )
        data.native_file = file
        session = get_session()
        if session:
            session.commit()

        data_path = geode_functions.data_file_path(data.id, data.native_file)
        os.makedirs(os.path.dirname(data_path), exist_ok=True)
        assert os.path.exists(data_path), f"File not found at {data_path}"
    response = client.post(route, json={"id": data.id})
    print(response.get_json())
    assert response.status_code == 200
    attributes = response.get_json()["attributes"]
    assert type(attributes) is list
    for attribute in attributes:
        assert "attribute_name" in attribute
        assert "min_value" in attribute
        assert "max_value" in attribute


def test_database_uri_path(client: FlaskClient) -> None:
    app = client.application
    with app.app_context():
        base_dir = os.path.abspath(os.path.dirname(__file__))
        expected_db_path = os.path.join(base_dir, "data", "project.db")
        expected_uri = f"sqlite:///{expected_db_path}"

        assert app.config["SQLALCHEMY_DATABASE_URI"] == expected_uri

        assert os.path.exists(expected_db_path)


def test_geode_object_inheritance(client: FlaskClient) -> None:
    route = "/opengeodeweb_back/geode_object_inheritance"
    # Test BRep
    response = client.post(route, json={"geode_object_type": "BRep"})
    assert response.status_code == 200
    json_data = response.get_json()
    parents = json_data["parents"]
    children = json_data["children"]
    assert "BRep" not in parents
    assert "BRep" not in children
    # Descendants
    assert "StructuralModel" in children
    assert "ImplicitStructuralModel" in children

    # Test CrossSection
    response = client.post(route, json={"geode_object_type": "CrossSection"})
    assert response.status_code == 200
    json_data = response.get_json()
    parents = json_data["parents"]
    children = json_data["children"]
    assert "CrossSection" not in parents
    assert "CrossSection" not in children
    # Parent
    assert "Section" in parents
    # Descendant
    assert "ImplicitCrossSection" in children

    # Test PolyhedralSolid3D
    response = client.post(route, json={"geode_object_type": "PolyhedralSolid3D"})
    assert response.status_code == 200
    json_data = response.get_json()
    parents = json_data["parents"]
    children = json_data["children"]
    assert "PolyhedralSolid3D" not in parents
    assert "PolyhedralSolid3D" not in children
    # Parent
    assert "VertexSet" in parents

    # Test all params
    def get_full_data() -> test_utils.JsonData:
        return {"geode_object_type": "BRep"}

    test_utils.test_route_wrong_params(client, route, get_full_data)


def test_model_components(client: FlaskClient) -> None:
    geode_object_type = "BRep"
    filename = "LS2.og_brep"
    response = test_save_viewable_file(client, geode_object_type, filename)
    assert response.status_code == 200
    assert "mesh_components" in response.get_json()
    mesh_components = response.get_json()["mesh_components"]
    assert isinstance(mesh_components, list)
    assert len(mesh_components) > 0
    name_is_uuid = False
    name_is_not_uuid = False
    for mesh_component in mesh_components:
        assert isinstance(mesh_component, object)
        assert isinstance(mesh_component["geode_id"], str)
        assert isinstance(mesh_component["viewer_id"], int)
        assert isinstance(mesh_component["name"], str)
        assert isinstance(mesh_component["type"], str)
        if mesh_component["name"] == mesh_component["geode_id"]:
            name_is_uuid = True
        else:
            name_is_not_uuid = True
        assert isinstance(mesh_component["boundaries"], list)
        for boundary_uuid in mesh_component["boundaries"]:
            assert isinstance(boundary_uuid, str)
        assert isinstance(mesh_component["internals"], list)
        for internal_uuid in mesh_component["internals"]:
            assert isinstance(internal_uuid, str)
    assert name_is_uuid is True
    assert name_is_not_uuid is True
    assert "collection_components" in response.get_json()
    collection_components = response.get_json()["collection_components"]
    assert isinstance(collection_components, list)
    for collection_component in collection_components:
        assert isinstance(collection_component, object)
        assert isinstance(collection_component["geode_id"], str)
        assert isinstance(collection_component["name"], str)
        assert isinstance(collection_component["items"], list)
        for item_uuid in collection_component["items"]:
            assert isinstance(item_uuid, str)


def test_export_project_route(client: FlaskClient, tmp_path: Path) -> None:
    route = "/opengeodeweb_back/export_project"
    snapshot = {
        "styles": {"1": {"visibility": True, "opacity": 1.0, "color": [0.2, 0.6, 0.9]}}
    }
    filename = "export_project_test.vease"
    project_folder = client.application.config["DATA_FOLDER_PATH"]
    os.makedirs(project_folder, exist_ok=True)
    database_root_path = os.path.join(project_folder, "project.db")
    with open(database_root_path, "wb") as f:
        f.write(b"test_project_db")

    with get_session() as session:
        session.query(Data).delete()
        session.commit()

        data1 = Data(
            id="test_data_1",
            geode_object="BRep",
            viewer_object="BRep",
            viewer_elements_type="default",
            native_file="native.txt",
        )
        data2 = Data(
            id="test_data_2",
            geode_object="Section",
            viewer_object="Section",
            viewer_elements_type="default",
            native_file="native.txt",
        )
        session.add(data1)
        session.add(data2)
        session.commit()

        data1_dir = os.path.join(project_folder, "test_data_1")
        os.makedirs(data1_dir, exist_ok=True)
        with open(os.path.join(data1_dir, "native.txt"), "w") as f:
            f.write("native file content")

        data2_dir = os.path.join(project_folder, "test_data_2")
        os.makedirs(data2_dir, exist_ok=True)
        with open(os.path.join(data2_dir, "native.txt"), "w") as f:
            f.write("native file content")

    response = client.post(route, json={"snapshot": snapshot, "filename": filename})
    assert response.status_code == 200
    assert response.headers.get("new-file-name") == filename
    assert response.mimetype == "application/octet-binary"
    response.direct_passthrough = False
    zip_bytes = response.get_data()
    tmp_zip_path = tmp_path / filename
    tmp_zip_path.write_bytes(zip_bytes)

    with zipfile.ZipFile(tmp_zip_path, "r") as zip_file:
        names = zip_file.namelist()
        assert "snapshot.json" in names
        parsed = json.loads(zip_file.read("snapshot.json").decode("utf-8"))
        assert parsed == snapshot
        assert "project.db" in names
        assert "test_data_1/native.txt" in names
        assert "test_data_2/native.txt" in names

    response.close()

    export_path = os.path.join(project_folder, filename)
    if os.path.exists(export_path):
        os.remove(export_path)


def test_import_project_route(client: FlaskClient, tmp_path: Path) -> None:
    route = "/opengeodeweb_back/import_project"
    snapshot = {
        "styles": {"1": {"visibility": True, "opacity": 1.0, "color": [0.2, 0.6, 0.9]}}
    }

    original_data_folder = client.application.config["DATA_FOLDER_PATH"]
    client.application.config["DATA_FOLDER_PATH"] = os.path.join(
        str(tmp_path), "project_data"
    )
    db_path = os.path.join(client.application.config["DATA_FOLDER_PATH"], "project.db")

    import sqlite3, zipfile, json

    temp_db = tmp_path / "temp_project.db"
    conn = sqlite3.connect(str(temp_db))
    conn.execute(
        "CREATE TABLE datas (id TEXT PRIMARY KEY, geode_object TEXT, viewer_object TEXT, viewer_elements_type TEXT, native_file TEXT, "
        "viewable_file TEXT, light_viewable_file TEXT)"
    )
    conn.commit()
    conn.close()

    z = tmp_path / "import_project_test.vease"
    with zipfile.ZipFile(z, "w", compression=zipfile.ZIP_DEFLATED) as zipf:
        zipf.writestr("snapshot.json", json.dumps(snapshot))
        zipf.write(str(temp_db), "project.db")

    with open(z, "rb") as f:
        resp = client.post(
            route,
            data={"file": (f, "import_project_test.vease")},
            content_type="multipart/form-data",
        )

    assert resp.status_code == 200
    assert resp.get_json().get("snapshot") == snapshot
    assert os.path.exists(db_path)

    from opengeodeweb_microservice.database import connection

    client.application.config["DATA_FOLDER_PATH"] = original_data_folder
    test_db_path = os.environ.get("TEST_DB_PATH")
    if test_db_path:
        connection.init_database(test_db_path, create_tables=True)

    client.application.config["DATA_FOLDER_PATH"] = original_data_folder


def test_save_viewable_workflow_from_object(client: FlaskClient) -> None:
    route = "/opengeodeweb_back/create/point"
    point_data = {
        "name": "workflow_point_3d",
        "x": 0.0,
        "y": 0.0,
        "z": 0.0,
    }

    response = client.post(route, json=point_data)
    assert response.status_code == 200

    data_id = response.get_json()["id"]
    assert isinstance(data_id, str) and len(data_id) > 0
    assert response.get_json()["geode_object_type"] == "PointSet3D"
    assert response.get_json()["viewable_file"].endswith(".vtp")



# Standard library imports
from __future__ import annotations
from abc import abstractmethod

# Third party imports
import opengeode as og
from opengeodeweb_microservice.database.data_types import ViewerType

# Local application imports
from .geode_object import GeodeObject


class GeodeMesh(GeodeObject):
    @classmethod
    def viewer_type(cls) -> ViewerType:
        return "mesh"

    @abstractmethod
    def vertex_attribute_manager(self) -> og.AttributeManager: ...



# Standard library imports
import os

# Third party imports
import flask
import opengeode
from opengeodeweb_microservice.schemas import get_schemas_dict

# Local application imports
from opengeodeweb_back import geode_functions, utils_functions
import opengeodeweb_back.routes.create.schemas as schemas
from opengeodeweb_back.geode_objects.geode_point_set3d import GeodePointSet3D
from opengeodeweb_back.geode_objects.geode_edged_curve3d import GeodeEdgedCurve3D

routes = flask.Blueprint("create", __name__, url_prefix="/create")
schemas_dict = get_schemas_dict(os.path.join(os.path.dirname(__file__), "schemas"))


@routes.route(
    schemas_dict["point"]["route"],
    methods=schemas_dict["point"]["methods"],
)
def point() -> flask.Response:
    """Endpoint to create a single point in 3D space."""
    json_data = utils_functions.validate_request(flask.request, schemas_dict["point"])
    params = schemas.Point.from_dict(json_data)

    # Create the point
    pointset = GeodePointSet3D()
    builder = pointset.builder()
    builder.set_name(params.name)
    builder.create_point(opengeode.Point3D([params.x, params.y, params.z]))

    # Save and get info
    result = utils_functions.generate_native_viewable_and_light_viewable_from_object(
        pointset
    )
    return flask.make_response(result, 200)



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import opengeode_geosciences as og_geosciences
import opengeode_inspector as og_inspector
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeMeshType

# Local application imports
from .geode_vertex_set import GeodeVertexSet


class GeodePointSet2D(GeodeVertexSet):
    point_set: og.PointSet2D

    def __init__(self, point_set: og.PointSet2D | None = None) -> None:
        self.point_set = point_set if point_set is not None else og.PointSet2D.create()
        super().__init__(self.point_set)

    @classmethod
    def geode_object_type(cls) -> GeodeMeshType:
        return "PointSet2D"

    def native_extension(self) -> str:
        return self.point_set.native_extension()

    @classmethod
    def is_3D(cls) -> bool:
        return False

    @classmethod
    def is_viewable(cls) -> bool:
        return True

    def builder(self) -> og.PointSetBuilder2D:
        return og.PointSetBuilder2D.create(self.point_set)

    @classmethod
    def load(cls, filename: str) -> GeodePointSet2D:
        return GeodePointSet2D(og.load_point_set2D(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.point_set_additional_files2D(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_point_set_loadable2D(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.PointSetInputFactory2D.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.PointSetOutputFactory2D.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.point_set_object_priority2D(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_point_set_saveable2D(self.point_set, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_point_set2D(self.point_set, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_point_set2D(
            self.point_set, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_point_set2D(
            self.point_set, filename_without_extension
        )

    def inspect(self) -> og_inspector.PointSetInspectionResult:
        return og_inspector.inspect_point_set2D(self.point_set)

    def assign_crs(
        self, crs_name: str, info: og_geosciences.GeographicCoordinateSystemInfo
    ) -> None:
        builder = self.builder()
        og_geosciences.assign_point_set_geographic_coordinate_system_info2D(
            self.point_set, builder, crs_name, info
        )

    def convert_crs(
        self, crs_name: str, info: og_geosciences.GeographicCoordinateSystemInfo
    ) -> None:
        builder = self.builder()
        og_geosciences.convert_point_set_coordinate_reference_system2D(
            self.point_set, builder, crs_name, info
        )

    def create_crs(
        self, crs_name: str, input: og.CoordinateSystem2D, output: og.CoordinateSystem2D
    ) -> None:
        builder = self.builder()
        og.create_point_set_coordinate_system2D(
            self.point_set, builder, crs_name, input, output
        )



from dataclasses_json import DataClassJsonMixin
from dataclasses import dataclass
from typing import Dict, Any


@dataclass
class ExportProject(DataClassJsonMixin):
    def __post_init__(self) -> None:
        print(self, flush=True)

    filename: str
    snapshot: Dict[str, Any]



from dataclasses_json import DataClassJsonMixin
from dataclasses import dataclass


@dataclass
class GeodeObjectInheritance(DataClassJsonMixin):
    def __post_init__(self) -> None:
        print(self, flush=True)

    geode_object_type: str



# Standard library imports
from __future__ import annotations
from abc import abstractmethod

# Third party imports
import opengeode as og
from opengeodeweb_microservice.database.data_types import ViewerType, ViewerElementsType

# Local application imports
from .geode_object import GeodeObject

ComponentRegistry = dict[og.ComponentType, list[og.uuid]]


class GeodeModel(GeodeObject):
    @classmethod
    def viewer_type(cls) -> ViewerType:
        return "model"

    @classmethod
    def viewer_elements_type(cls) -> ViewerElementsType:
        return "default"

    @abstractmethod
    def mesh_components(self) -> ComponentRegistry: ...

    @abstractmethod
    def collection_components(self) -> ComponentRegistry: ...

    @abstractmethod
    def boundaries(self, id: og.uuid) -> list[og.ComponentID]: ...

    @abstractmethod
    def internals(self, id: og.uuid) -> list[og.ComponentID]: ...

    @abstractmethod
    def items(self, id: og.uuid) -> list[og.ComponentID]: ...

    @abstractmethod
    def component_name(self, id: og.uuid) -> str | None: ...



# Standard library imports
from typing import Callable, Any

# Third party imports
from flask.testing import FlaskClient

# Local application imports

JsonData = dict[str, Any]


def test_route_wrong_params(
    client: FlaskClient,
    route: str,
    get_full_data: Callable[[], JsonData],
    path: list[str | int] | None = None,
) -> None:
    if path is None:
        path = []

    def get_json() -> tuple[JsonData, Any]:
        json = get_full_data()
        target: Any = json
        for sub_path in path:
            target = target[sub_path]
        return json, target

    data = get_json()[1]
    if isinstance(data, dict):
        for key, value in data.items():
            json, target = get_json()
            target.pop(key)
            response = client.post(route, json=json)
            if response.status_code == 400:
                error_description: str = response.get_json()["description"]
                assert "must contain" in error_description
                assert f"'{key}'" in error_description
            if isinstance(value, (dict, list)):
                test_route_wrong_params(client, route, get_full_data, path + [key])

        json, target = get_json()
        target["dumb_key"] = "dumb_value"
        response = client.post(route, json=json)
        assert response.status_code == 400
        error_description = response.get_json()["description"]
        assert "must not contain" in error_description
        assert "'dumb_key'" in error_description
    elif isinstance(data, list) and data:
        test_route_wrong_params(client, route, get_full_data, path + [0])



# Standard library imports
import os

# Third party imports
import werkzeug
import flask
from opengeodeweb_microservice.database.data import Data
from opengeodeweb_microservice.database.data_types import (
    GeodeObjectType,
    geode_object_type,
)

# Local application imports
from .geode_objects import geode_objects
from .geode_objects.geode_object import GeodeObject


def data_file_path(data_id: str, filename: str | None = None) -> str:
    data_folder_path = flask.current_app.config["DATA_FOLDER_PATH"]
    data_path = os.path.join(data_folder_path, data_id)
    if filename is not None:
        return os.path.join(data_path, filename)
    return data_path


def geode_object_from_string(value: str) -> type[GeodeObject]:
    return geode_objects[geode_object_type(value)]


def load_geode_object(data_id: str) -> GeodeObject:
    data = Data.get(data_id)
    if not data:
        flask.abort(404, f"Data with id {data_id} not found")

    file_absolute_path = data_file_path(data_id, data.native_file)
    print("Loading file: ", file_absolute_path)
    print("File exists: ", os.path.exists(file_absolute_path))
    return geode_object_from_string(data.geode_object).load(file_absolute_path)


def get_data_info(data_id: str) -> Data:
    data = Data.get(data_id)
    if not data:
        flask.abort(404, f"Data with id {data_id} not found")
    return data


def upload_file_path(filename: str) -> str:
    upload_folder = flask.current_app.config["UPLOAD_FOLDER"]
    secure_filename = werkzeug.utils.secure_filename(filename)
    return os.path.abspath(os.path.join(upload_folder, secure_filename))


def geode_object_output_extensions(
    geode_object: GeodeObject,
) -> dict[GeodeObjectType, dict[str, bool]]:
    results: dict[GeodeObjectType, dict[str, bool]] = {}
    for mixin_geode_object in geode_objects[geode_object.geode_object_type()].__mro__:
        output_extensions_method = getattr(
            mixin_geode_object, "output_extensions", None
        )
        if output_extensions_method is None:
            continue
        output_extensions = output_extensions_method.__func__(mixin_geode_object)
        if output_extensions is None:
            continue
        object_output_extensions: dict[str, bool] = {}
        is_saveable_method = getattr(mixin_geode_object, "is_saveable")
        for output_extension in output_extensions:
            bool_is_saveable = is_saveable_method(
                geode_object, f"test.{output_extension}"
            )
            object_output_extensions[output_extension] = bool_is_saveable
        if hasattr(mixin_geode_object, "geode_object_type"):
            results[mixin_geode_object.geode_object_type()] = object_output_extensions
    return results



from dataclasses_json import DataClassJsonMixin
from dataclasses import dataclass
from typing import Optional


@dataclass
class UploadFile(DataClassJsonMixin):
    def __post_init__(self) -> None:
        print(self, flush=True)

    filename: Optional[str] = None



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import opengeode_geosciences as og_geosciences
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeModelType

# Local application imports
from .geode_brep import GeodeBRep


class GeodeStructuralModel(GeodeBRep):
    structural_model: og_geosciences.StructuralModel

    def __init__(
        self, structural_model: og_geosciences.StructuralModel | None = None
    ) -> None:
        self.structural_model = (
            structural_model
            if structural_model is not None
            else og_geosciences.StructuralModel()
        )
        super().__init__(self.structural_model)

    @classmethod
    def geode_object_type(cls) -> GeodeModelType:
        return "StructuralModel"

    def native_extension(self) -> str:
        return self.structural_model.native_extension()

    def builder(self) -> og_geosciences.StructuralModelBuilder:
        return og_geosciences.StructuralModelBuilder(self.structural_model)

    @classmethod
    def load(cls, filename: str) -> GeodeStructuralModel:
        return GeodeStructuralModel(og_geosciences.load_structural_model(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og_geosciences.structural_model_additional_files(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og_geosciences.is_structural_model_loadable(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og_geosciences.StructuralModelInputFactory.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og_geosciences.StructuralModelOutputFactory.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og_geosciences.structural_model_object_priority(filename)

    def is_saveable(self, filename: str) -> bool:
        return og_geosciences.is_structural_model_saveable(
            self.structural_model, filename
        )

    def save(self, filename: str) -> list[str]:
        return og_geosciences.save_structural_model(self.structural_model, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_structural_model(
            self.structural_model, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_structural_model(
            self.structural_model, filename_without_extension
        )



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeMeshType

# Local application imports
from .geode_surface_mesh2d import GeodeSurfaceMesh2D


class GeodePolygonalSurface2D(GeodeSurfaceMesh2D):
    polygonal_surface: og.PolygonalSurface2D

    def __init__(self, polygonal_surface: og.PolygonalSurface2D | None = None) -> None:
        self.polygonal_surface = (
            polygonal_surface
            if polygonal_surface is not None
            else og.PolygonalSurface2D.create()
        )
        super().__init__(self.polygonal_surface)

    @classmethod
    def geode_object_type(cls) -> GeodeMeshType:
        return "PolygonalSurface2D"

    def native_extension(self) -> str:
        return self.polygonal_surface.native_extension()

    def builder(self) -> og.PolygonalSurfaceBuilder2D:
        return og.PolygonalSurfaceBuilder2D.create(self.polygonal_surface)

    @classmethod
    def load(cls, filename: str) -> GeodePolygonalSurface2D:
        return GeodePolygonalSurface2D(og.load_polygonal_surface2D(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.polygonal_surface_additional_files2D(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_polygonal_surface_loadable2D(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.PolygonalSurfaceInputFactory2D.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.PolygonalSurfaceOutputFactory2D.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.polygonal_surface_object_priority2D(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_polygonal_surface_saveable2D(self.polygonal_surface, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_polygonal_surface2D(self.polygonal_surface, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_polygonal_surface2D(
            self.polygonal_surface, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_polygonal_surface2D(
            self.polygonal_surface, filename_without_extension
        )



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeMeshType

# Local application imports
from .geode_grid2d import GeodeGrid2D


class GeodeLightRegularGrid2D(GeodeGrid2D):
    light_regular_grid: og.LightRegularGrid2D

    def __init__(self, light_regular_grid: og.LightRegularGrid2D) -> None:
        self.light_regular_grid = light_regular_grid
        super().__init__(self.light_regular_grid)

    @classmethod
    def geode_object_type(cls) -> GeodeMeshType:
        return "LightRegularGrid2D"

    def native_extension(self) -> str:
        return self.light_regular_grid.native_extension()

    def builder(self) -> og.IdentifierBuilder:
        return og.IdentifierBuilder(self.light_regular_grid)

    @classmethod
    def load(cls, filename: str) -> GeodeLightRegularGrid2D:
        return GeodeLightRegularGrid2D(og.load_light_regular_grid2D(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.light_regular_grid_additional_files2D(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_light_regular_grid_loadable2D(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.LightRegularGridInputFactory2D.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.LightRegularGridOutputFactory2D.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.light_regular_grid_object_priority2D(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_light_regular_grid_saveable2D(self.light_regular_grid, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_light_regular_grid2D(self.light_regular_grid, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_light_regular_grid2D(
            self.light_regular_grid, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_light_regular_grid2D(
            self.light_regular_grid, filename_without_extension
        )

    def vertex_attribute_manager(self) -> og.AttributeManager:
        return self.light_regular_grid.grid_vertex_attribute_manager()

    def cell_attribute_manager(self) -> og.AttributeManager:
        return self.light_regular_grid.cell_attribute_manager()



# Standard library imports
import uuid

# Third party imports
import pytest
from flask.testing import FlaskClient

# Local application imports
from opengeodeweb_back import test_utils


@pytest.fixture
def point_data() -> test_utils.JsonData:
    return {"name": "test_point", "x": 1.0, "y": 2.0, "z": 3.0}


def test_create_point(client: FlaskClient, point_data: test_utils.JsonData) -> None:
    """Test the creation of a point with valid data."""
    route: str = "/opengeodeweb_back/create/point"

    # Test with all required data
    response = client.post(route, json=point_data)
    assert response.status_code == 200

    # Verify response data
    response_data = response.get_json()
    assert "viewable_file" in response_data
    assert "id" in response_data
    assert "name" in response_data
    assert "native_file" in response_data
    assert "viewer_type" in response_data
    assert "geode_object_type" in response_data

    assert response_data["name"] == point_data["name"]
    assert response_data["viewer_type"] == "mesh"
    assert response_data["geode_object_type"] == "PointSet3D"

    # Test with missing parameters
    test_utils.test_route_wrong_params(client, route, lambda: point_data.copy())


def test_create_point_with_invalid_data(client: FlaskClient) -> None:
    """Test the point creation endpoint with invalid data."""
    route: str = "/opengeodeweb_back/create/point"

    # Test with non-numeric coordinates
    invalid_data: test_utils.JsonData = {
        "name": "invalid_point",
        "x": "not_a_number",
        "y": 2.0,
        "z": 3.0,
    }
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400

    # Test with missing coordinates
    invalid_data = {"name": "invalid_point", "y": 2.0, "z": 3.0}
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400



# Standard library imports
from __future__ import annotations

# Third party imports
from opengeodeweb_microservice.database.data_types import GeodeObjectType

# Local application imports
from .geode_object import GeodeObject
from .geode_brep import GeodeBRep
from .geode_vertex_set import GeodeVertexSet
from .geode_graph import GeodeGraph
from .geode_point_set2d import GeodePointSet2D
from .geode_point_set3d import GeodePointSet3D
from .geode_edged_curve2d import GeodeEdgedCurve2D
from .geode_edged_curve3d import GeodeEdgedCurve3D
from .geode_raster_image2d import GeodeRasterImage2D
from .geode_raster_image3d import GeodeRasterImage3D
from .geode_polygonal_surface2d import GeodePolygonalSurface2D
from .geode_polygonal_surface3d import GeodePolygonalSurface3D
from .geode_triangulated_surface2d import GeodeTriangulatedSurface2D
from .geode_triangulated_surface3d import GeodeTriangulatedSurface3D
from .geode_regular_grid2d import GeodeRegularGrid2D
from .geode_polyhedral_solid3d import GeodePolyhedralSolid3D
from .geode_tetrahedral_solid3d import GeodeTetrahedralSolid3D
from .geode_hybrid_solid3d import GeodeHybridSolid3D
from .geode_regular_grid3d import GeodeRegularGrid3D
from .geode_light_regular_grid2d import GeodeLightRegularGrid2D
from .geode_light_regular_grid3d import GeodeLightRegularGrid3D
from .geode_section import GeodeSection
from .geode_structural_model import GeodeStructuralModel
from .geode_cross_section import GeodeCrossSection
from .geode_implicit_structural_model import GeodeImplicitStructuralModel
from .geode_implicit_cross_section import GeodeImplicitCrossSection

geode_objects: dict[GeodeObjectType, type[GeodeObject]] = {
    "VertexSet": GeodeVertexSet,
    "Graph": GeodeGraph,
    "PointSet2D": GeodePointSet2D,
    "PointSet3D": GeodePointSet3D,
    "EdgedCurve2D": GeodeEdgedCurve2D,
    "EdgedCurve3D": GeodeEdgedCurve3D,
    "RasterImage2D": GeodeRasterImage2D,
    "RasterImage3D": GeodeRasterImage3D,
    "PolygonalSurface2D": GeodePolygonalSurface2D,
    "PolygonalSurface3D": GeodePolygonalSurface3D,
    "TriangulatedSurface2D": GeodeTriangulatedSurface2D,
    "TriangulatedSurface3D": GeodeTriangulatedSurface3D,
    "RegularGrid2D": GeodeRegularGrid2D,
    "PolyhedralSolid3D": GeodePolyhedralSolid3D,
    "TetrahedralSolid3D": GeodeTetrahedralSolid3D,
    "HybridSolid3D": GeodeHybridSolid3D,
    "RegularGrid3D": GeodeRegularGrid3D,
    "LightRegularGrid2D": GeodeLightRegularGrid2D,
    "LightRegularGrid3D": GeodeLightRegularGrid3D,
    "BRep": GeodeBRep,
    "Section": GeodeSection,
    "StructuralModel": GeodeStructuralModel,
    "CrossSection": GeodeCrossSection,
    "ImplicitStructuralModel": GeodeImplicitStructuralModel,
    "ImplicitCrossSection": GeodeImplicitCrossSection,
}



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
from opengeodeweb_microservice.database.data_types import (
    GeodeMeshType,
    ViewerElementsType,
)

# Local application imports
from .geode_mesh import GeodeMesh


class GeodeVertexSet(GeodeMesh):
    vertex_set: og.VertexSet

    def __init__(self, vertex_set: og.VertexSet | None = None) -> None:
        self.vertex_set = (
            vertex_set if vertex_set is not None else og.VertexSet.create()
        )
        super().__init__(self.vertex_set)

    @classmethod
    def geode_object_type(cls) -> GeodeMeshType:
        return "VertexSet"

    @classmethod
    def viewer_elements_type(cls) -> ViewerElementsType:
        return "points"

    def native_extension(self) -> str:
        return self.vertex_set.native_extension()

    @classmethod
    def is_3D(cls) -> bool:
        return False

    @classmethod
    def is_viewable(cls) -> bool:
        return False

    def builder(self) -> og.VertexSetBuilder:
        return og.VertexSetBuilder.create(self.vertex_set)

    @classmethod
    def load(cls, filename: str) -> GeodeVertexSet:
        return GeodeVertexSet(og.load_vertex_set(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.vertex_set_additional_files(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_vertex_set_loadable(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.VertexSetInputFactory.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.VertexSetOutputFactory.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.vertex_set_object_priority(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_vertex_set_saveable(self.vertex_set, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_vertex_set(self.vertex_set, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return ""

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return ""

    def inspect(self) -> object:
        return None

    def vertex_attribute_manager(self) -> og.AttributeManager:
        return self.vertex_set.vertex_attribute_manager()
